/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/***									***/ 
/*** 		Authors : Dan Kershaw & Gary Cook			***/
/*** 		Cambridge University Engineering Department		***/
/*** 		Trumpington Street, Cambridge, CB2 1PZ, England.	***/
/***									***/
/*** 		getArgs.cc : Source for the command line processing	***/
/***									***/
/***		=======   COPYRIGHT NOTICE   =======			***/
/***		Copyright (c) 1996, Dan Kerhaw & Gary Cook		***/
/*** 		All rights reserved.					***/
/***									***/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include "getArgs.h"

/**********************************************/
/* PRIVATE FUNCTIONS                          */
/**********************************************/

/*--------------------------------------------*/
/* FUNCTION -> updateArgs                     */
/* Remove parsed arguments from the argument  */
/* list, and also updates the number.         */
/*--------------------------------------------*/
void getArgs::updateArgs( int rm_cnt ) {
                      
  /* update the argument count */
  (*argc)-- ;
 
  /* update the command line */
  for( int i = rm_cnt ; i < *argc ; i++ ) 
    argv[i] = argv[i+1] ;

}

/*--------------------------------------------*/
/* Function -> checkArgError                  */
/* Check for a valid integer argument         */
/*--------------------------------------------*/
bool getArgs::checkArgError( char *argument,
			     char *flag, 
			     int defValue, 
			     char *helpMsg ) {

  bool errorFlag = false;

  // allocate space and a stream for an error message
  char *errMsg = new char[LINESIZE];
  // ostrstream errMsgStr(errMsg, LINESIZE);

  // check that the argument is a valid one (ie an integer)
  int length = strlen(argument);
  for (int i = 0; i < length; i++)
    if ( (!isdigit(argument[i]) && argument[0] != '-' && i == 0 ) ||
	 (!isdigit(argument[i]) && i != 0 ) ) {
      // errMsgStr << "Flag " << flag << ": Argument (";
      // errMsgStr << argument << ") must be an INTEGER!\n"  << '\0';
      sprintf(errMsg, "Flag %s: Argument (%s) must be an INTEGER!\n",
	      flag, argument);
      success = false;
      errorFlag = true;
      break;
    }

  if (!errorFlag)
    errMsg = NULL;

  // convert int to char to pass default value to option info
  char *defchar = new char[WORDSIZE];
  // ostrstream defcharStr(defchar, WORDSIZE);
  // defcharStr << defValue  << '\0';
  sprintf(defchar, "%d", defValue);
  addOptionInfo( flag, defchar, helpMsg, errMsg );
  
  return errorFlag;

}

/*--------------------------------------------*/
/* FUNCTION -> checkArgError                  */
/* Check for a valid floating argument        */
/*--------------------------------------------*/
bool getArgs::checkArgError( char *argument,
			     char *flag, 
			     float defValue, 
			     char *helpMsg ) {

  bool errorFlag = false;

  // allocate space and a stream for an error message
  char *errMsg = new char[LINESIZE];
  // ostrstream errMsgStr(errMsg, LINESIZE);

  // check that the argument is a valid one (ie an integer)
  int length = strlen(argument);
  for (int i = 0; i < length; i++)
    if ( ( !isdigit(argument[i]) && 
	   argument[0] != '-' && argument[0] != '.' && i == 0 ) ||
	 (!isdigit(argument[i]) && argument[i] != '.' && i != 0 ) ) {
      // errMsgStr << "Flag " << flag << ": Argument (";
      // errMsgStr << argument << ") must be a FLOAT!\n"  << '\0';
      sprintf(errMsg, "Flag %s : Argument (%s) must be a FLOAT!\n",
	      flag, argument);
      success = false;
      errorFlag = true;
      break;
    }
 
  if (!errorFlag)
    errMsg = NULL;

  // convert int to char to pass default value to option info
  char *defchar = new char[WORDSIZE];
  // ostrstream defcharStr(defchar, WORDSIZE);
  // defcharStr << defValue << '\0';
  sprintf(defchar, "%f", defValue);
  addOptionInfo( flag, defchar, helpMsg, errMsg );
  
  return errorFlag;

}

/*--------------------------------------------*/
/* FUNCTION -> checkArgError                  */
/* Check for a valid string argument          */
/*--------------------------------------------*/
bool getArgs::checkArgError( char *argument,
			     char *flag, 
			     char *defValue, 
			     char *helpMsg ) {

  bool errorFlag = false;

  // allocate space and a stream for an error message
  char *errMsg = new char[LINESIZE];
  // ostrstream errMsgStr(errMsg, LINESIZE);

  // check that the argument is a valid one (ie an integer)
  int length = strlen(argument);
  for (int i = 0; i < length; i++)
    if ( !isgraph(argument[i] )) {
      // errMsgStr << "Flag " << flag << ": Argument (";
      // errMsgStr << argument << ") must be a String!\n" << '\0';
      sprintf(errMsg, "Flag %s : Argument (%s) must be a STRING!\n",
	      flag, argument);
      success = false;
      errorFlag = true;
      break;
    }

  if (!errorFlag)
    errMsg = NULL;
  
  addOptionInfo( flag, defValue, helpMsg, errMsg );
  return errorFlag;

}

/*--------------------------------------------*/
/* FUNCTION -> checkArgError                  */
/* Check for a valid flag                     */
/*--------------------------------------------*/
bool getArgs::checkArgError( char *flag, 
			     bool defValue, 
			     char *helpMsg ) {
  
  if (defValue)
    addOptionInfo( flag, "true", helpMsg );
  else
    addOptionInfo( flag, "false", helpMsg );

  return false;

}

/*--------------------------------------------*/
/* FUNCTION -> addOptionInfo                  */
/* Add option information to a linked list    */
/* Note that errMsg is optional               */
/*--------------------------------------------*/
void getArgs::addOptionInfo( char *flag, 
			     char *defValue, 
			     char *help, 
			     char *errMsg ) {

  OptInfo *opt = optInfo;

  // find the end of the linked list
  for ( ; opt->next != NULL; opt = opt->next );

  // set the info
  opt->flag = flag;
  opt->help = help;
  opt->defValue = strdup(defValue);
  if (errMsg != NULL)
    opt->errMsg = strdup(errMsg);
  else
    opt->errMsg = NULL;
  
  // create next link and set the pointer to the next link equal to NULL
  opt->next = new OptInfo[1];
  opt = opt->next;
  opt->next = NULL;

}

/*--------------------------------------------*/
/* FUNCTION -> printErrors                    */
/* Print the errors found when parsing the    */
/* command line                               */
/*--------------------------------------------*/
void getArgs::printErrors ( void ) {

  fprintf(stderr, "Error with the command-line options!\n") ;
  for ( OptInfo *opt = optInfo; opt->next != NULL; opt = opt->next )
    if (opt->errMsg != NULL)
      fprintf(stderr, "%s\n", opt->errMsg) ;
}

/*--------------------------------------------*/
/* FUNCTION -> printHelp                      */
/* Print the help messages                    */
/*--------------------------------------------*/
void getArgs::printHelp ( void ) {

  fprintf(stderr, "\nOPTIONS for %s :\nSwitches\t\tInfo\t\t\t\t     Default\n",
	  argv[0]) ;

  for ( OptInfo *opt = optInfo; opt->next != NULL; opt = opt->next )
    if (opt->help != NULL)
      fprintf(stderr, "%s\t\t%s\t\t%s\n", opt->flag, opt->help, opt->defValue);
    else
      fprintf(stderr, "%s\t\t\t\t\t\t\t\t%s\n", opt->flag, opt->defValue);

  exit(1);

}
    
  
/**********************************************/
/* PUBLIC FUNCTIONS                           */
/**********************************************/

/*--------------------------------------------*/
/* FUNCTION -> intArg                         */
/* Parse an integer argument                  */
/*--------------------------------------------*/
int getArgs::intArg( char *flag, 
		     char *helpMsg,
		     int defValue ) {

  int value = defValue;

  for(int i = 1; i < *argc - 1; i++)
    if(!strcmp(argv[i], flag)) {
      if (checkArgError(argv[i+1], flag, defValue, helpMsg))
	updateArgs(i);
      else {
	value = atoi(argv[i + 1]);
	updateArgs(i+1);
	updateArgs(i);
      }
      return(value);
    }
  
  // if flag not found use checkArgError to add option info
  checkArgError("0", flag, defValue, helpMsg) ;
  return(value);
}

/*--------------------------------------------*/
/* FUNCTION -> intArg                         */
/* Parse an integer argument                  */
/*--------------------------------------------*/
int getArgs::intArg( char *flag, 
		     int defValue ) {

  int value = defValue;

  for(int i = 1; i < *argc - 1; i++)
    if(!strcmp(argv[i], flag)) {
      if (checkArgError(argv[i+1], flag, defValue, NULL))
	updateArgs(i);
      else {
	value = atoi(argv[i + 1]);
	updateArgs(i+1);
	updateArgs(i);
      }
      return(value);
    }
  
  // if flag not found use checkArgError to add option info
  checkArgError("0", flag, defValue, NULL) ;
  return(value);
}

/*--------------------------------------------*/
/* FUNCTION -> floatArg                       */
/* Parse an float argument                    */
/*--------------------------------------------*/
float getArgs::floatArg( char *flag,
			 char *helpMsg,
			 float defValue ) {
  
  float value = defValue;

  
  for(int i = 1; i < *argc - 1; i++)
    if(!strcmp(argv[i], flag)) {
      if (checkArgError(argv[i+1], flag, defValue, helpMsg))
	updateArgs(i);
      else {
	value = (float) atof(argv[i + 1]);
	updateArgs(i+1);
	updateArgs(i);
      }
      return(value);
    }
  
  // if flag not found use checkArgError to add option info
  checkArgError("0", flag, defValue, helpMsg) ;
  return(value);
  
}

/*--------------------------------------------*/
/* FUNCTION -> floatArg                       */
/* Parse an float argument                    */
/*--------------------------------------------*/
float getArgs::floatArg( char *flag,
			 float defValue ) {
  
  float value = defValue;

  
  for(int i = 1; i < *argc - 1; i++)
    if(!strcmp(argv[i], flag)) {
      if (checkArgError(argv[i+1], flag, defValue, NULL))
	updateArgs(i);
      else {
	value = (float) atof(argv[i + 1]);
	updateArgs(i+1);
	updateArgs(i);
      }
      return(value);
    }
  
  // if flag not found use checkArgError to add option info
  checkArgError("0", flag, defValue, NULL) ;
  return(value);
  
}

/*--------------------------------------------*/
/* FUNCTION -> stringArg                      */
/* Parse an double argument                   */
/*--------------------------------------------*/
char *getArgs::stringArg( char *flag, 
			  char *helpMsg,
			  char *defValue ) {

  char *value = defValue;

  for(int i = 1; i < *argc - 1; i++)
    if(!strcmp(argv[i], flag)) {
       if (checkArgError(argv[i+1], flag, defValue, helpMsg))
	updateArgs(i);
      else {
	value = argv[i + 1];
	updateArgs(i+1);
	updateArgs(i);
      }
      return(value);
    }
  
  // if flag not found use checkArgError to add option info
  checkArgError("None", flag, defValue, helpMsg) ;
  return(value);
  
}  

/*--------------------------------------------*/
/* FUNCTION -> stringArg                      */
/* Parse an double argument                   */
/*--------------------------------------------*/
char *getArgs::stringArg( char *flag, 
			  char *defValue ) {

  char *value = defValue;

  for(int i = 1; i < *argc - 1; i++)
    if(!strcmp(argv[i], flag)) {
       if (checkArgError(argv[i+1], flag, defValue, NULL))
	updateArgs(i);
      else {
	value = argv[i + 1];
	updateArgs(i+1);
	updateArgs(i);
      }
      return(value);
    }
  
  // if flag not found use checkArgError to add option info
  checkArgError("None", flag, defValue, NULL) ;
  return(value);
  
}  
  
/*--------------------------------------------*/
/* FUNCTION -> switchArg                      */
/* Parse an double argument                   */
/*--------------------------------------------*/
bool getArgs::switchArg( char *flag, 
			 char *helpMsg,
			 bool defValue ) {

  bool value = defValue;

  for(int i = 1; i < *argc; i++)
    if(!strcmp(argv[i], flag)) {
      value = true;
      updateArgs(i);
      checkArgError(flag, defValue, helpMsg) ;
      return(value);
    }
  
  // if flag not found use checkArgError to add option info
  checkArgError(flag, defValue, helpMsg) ;
  return(value);
  
}  
  
/*--------------------------------------------*/
/* FUNCTION -> switchArg                      */
/* Parse an double argument                   */
/*--------------------------------------------*/
bool getArgs::switchArg( char *flag, 
			 bool defValue ) {

  bool value = defValue;

  for(int i = 1; i < *argc; i++)
    if(!strcmp(argv[i], flag)) {
      value = true;
      updateArgs(i);
      checkArgError(flag, defValue, NULL) ;
      return(value);
    }
  
  // if flag not found use checkArgError to add option info
  checkArgError(flag, defValue, NULL) ;
  return(value);
  
}  
  

/*--------------------------------------------*/
/* FUNCTION -> printUsage                     */
/* Do as it says                              */
/*--------------------------------------------*/
void getArgs::printUsage( char *firstline ) {

  if (firstline != NULL)
    fprintf(stderr, "USAGE: ");
  printHelp();

}

/*--------------------------------------------*/
/* FUNCTION -> finalCheck                     */
/* Do a final check to see if error messages  */
/* must be printed along with usage.          */
/*--------------------------------------------*/
void getArgs::finalCheck( int numReqArgs,
			  char *firstline ) {

  // do a check for extemporaneous options
  if (*argc != numReqArgs + 1) {
    for (int i = 1; i < *argc; i++)
      if (strlen(argv[i]) > 1 && argv[i][0] == '-')
	fprintf(stderr, "Unknown or illegal flag used! %s not recognised\n", 
		argv[i] ) ;
    success = false;
  }

  if ( !success ) {
    printErrors();
    if (firstline != NULL)
      fprintf(stderr, "USAGE: %s", firstline);
    printHelp();
  }

}
