#include "incs.h"
#include "lu.h"
#include "utils.h"
#include "matrix.h"

void lin_newton(real *grad,real **hess, real *theta,int ipdim,float lambda) {
     int bad_flag;
     real *new_theta;
     int singular;
     long d;
     
     new_theta = RealVector(ipdim);
     
     if(issingular(hess,ipdim)) 
       {
	    print_matrix(hess,ipdim,ipdim);
	    error(1,"Singular matrix in newton");
	    singular = 1;  
       }
     
     copy_vector(theta,new_theta,ipdim);
     
     lin_solve(hess,grad,ipdim,theta);
     
     for(d=0;d<ipdim;d++)
       {
	    theta[d] *= lambda;
	    new_theta[d] += theta[d];
       }
     
     
     bad_flag = 0;
     if(!bad_flag)
       for(d=0;d<ipdim;d++)
	 if(isnan(new_theta[d]))
	   {
		error(1,"theta is nan");
		bad_flag = 1;
	   }
     if(!bad_flag) 
       {
	    for(d=0;d<ipdim;d++)
	      theta[d] = new_theta[d];
       }
     FreeRealVector(&new_theta);
}
 



/* linSolve:  Solve for a in a X = b given X and b */

void lin_solve(real **matr, real *vect, int ndim, real *answ) {
  real **lud = RealMatrix2(ndim, ndim), d;
  int   *indx = IntVector(ndim);

  lu_decomp(matr, lud, ndim, indx, &d);
  lu_back_sub(vect, lud, indx, ndim, answ);

  free(lud);
  free(indx);
}


/* functionally equivalent to the Numerical Recipes ludcmp()
   Buy the book! */

void lu_decomp(real** mat, real** a, int ndim, int* indx, real* d) {
  int i, imax = -100000, j ,k;
  real big, sum;
  real *vv = RealVector(ndim);

  /* copy input so remainder can operate in place */
  matCopy(mat, a, ndim, ndim);

  *d = 1.0;
  for(i = 0; i < ndim; i++) {
    big = 0.0;
    for(j = 0; j < ndim; j++)
      if(fabs(a[i][j]) > big)
	big = fabs(a[i][j]);
    if(big == 0.0) error(1,"luDecomp: Singular matrix\n");
    vv[i] = 1.0 / big;
  }
  for(j = 0; j < ndim; j++) {
    for(i = 0; i < j; i++) {
      sum = a[i][j];
      for(k = 0; k < i; k++) sum -= a[i][k] * a[k][j];
      a[i][j] = sum;
    }
    big = 0.0;
    for(i = j; i < ndim; i++) {
      sum = a[i][j];
      for(k = 0; k < j; k++)
	sum -= a[i][k] * a[k][j];
      a[i][j] = sum;
      if(vv[i] * fabs(sum) >= big) {
	big  = vv[i] * fabs(sum);
	imax = i;
      }
    }
    if(j != imax) {
      for(k = 0; k < ndim; k++) {
	real swap = a[imax][k];
	a[imax][k] = a[j][k];
	a[j][k] = swap;
      }
      *d = -(*d);
      vv[imax] = vv[j];
    }
    indx[j] = imax;
    if(a[j][j] == 0.0) a[j][j] = TINY_FLOAT;
    if(j != ndim) {
      real scale = 1.0 / a[j][j];
      for(i = j + 1; i < ndim; i++)
	a[i][j] *= scale;
    }
  }
  free(vv);
}

/* functionally equivalent to the Numerical Recipes lubksb()
   Buy the book! */

void lu_back_sub(real *vect, real **lud, int *indx, int ndim, real *answ) {
  int i, ii = -1, ip, j;
  real sum;

  /* copy input so remainder can operate in place */
  for(i = 0; i < ndim; i++)
    answ[i] = vect[i];
  
  for(i = 0; i < ndim; i++) {
    ip = indx[i];
    sum = answ[ip];
    answ[ip] = answ[i];
    if(ii != -1)
      for(j = ii; j < i; j++) sum -= lud[i][j] * answ[j];
    else if(sum != 0.0) ii = i;
    answ[i] = sum;
  }
  for(i = ndim - 1; i >= 0; i--) {
    sum = answ[i];
    for(j = i + 1; j < ndim; j++) sum -= lud[i][j] * answ[j];
    answ[i] = sum / lud[i][i];
  }
}


/* matInverse:  compute matrix inverse returning the determinant */

real lu_inverse(real** orig, real** invr, int ndim) {
  real **lud = RealMatrix2(ndim, ndim), d;
  int   *indx = IntVector(ndim);
  real *unitVec = RealVector(ndim);
  real *column = RealVector(ndim);
  real det = 0.0;
  int   i, j;

  for(i = 0; i < ndim; i++)
    unitVec[i] = 0.0;

  lu_decomp(orig, lud, ndim, indx, &d);
  for(j = 0; j < ndim; j++) {
	det += log(fabs(lud[j][j]));
  }
  for(j = 0; j < ndim; j++) {
    unitVec[j] = 1.0;
    lu_back_sub(unitVec, lud, indx, ndim, column);
    unitVec[j] = 0.0;
    for(i = 0; i < ndim; i++)
      invr[i][j] = column[i];
  }

  free(lud);
  free(indx);

  return(det);
}


