#include "utils.h"
#include "consts.h"
#include "chol.h"
#include "matrix.h"
#include <math.h>

#define TINY 1.0E-20
#define MIN_SCALE 1.0E-300

int isposdef(real** matrix, int dim) {
  int pos_def;
  real **decomp = RealMatrix2(dim, dim);
  return pos_def = cholesky_decomp(matrix, decomp, dim);
  FreeRealMatrix2(&decomp);
}


int issingular(real **matrix,int n) {
	int i,j;
	int singular;
	real big;
	big = TINY;
	for(i=0;i<n;i++) 
		for(j=0;j<n;j++)
			if(fabs(matrix[j][i]) > big)
				big = fabs(matrix[j][i]);
	if(big < TINY)
		singular = 1; 
	
	else singular = 0;
	for(i=0;i<n;i++) 
		for(j=0;j<n;j++)
			if(fabs(matrix[i][j]) > big)
				big = fabs(matrix[i][j]);
	if(big < TINY)
		singular = 1; 
	else singular = 0;  
	return singular;
}


real log_trace(real **mat, int n)
{
	int i;
	real trace = 0.0;
	
	for(i = 0; i < n; i ++)
	{
		if(mat[i][i] < 0.0) error(1,"matrix diagonal < 0.0");
		if(mat[i][i] == 0.0) trace += LZERO;
		else                 trace += log(mat[i][i]);
	}
	return trace;
}

real invert_matrix(real **mat, real **inv, int dim)
{
    long i, j, k;
    real pos_def;
    real **decomp = RealMatrix2(dim, dim);
    real sum;
	real det;

    pos_def = cholesky_decomp(mat, decomp, dim);

    if(pos_def == 0) 
	{
		FreeRealMatrix2(&decomp);
        error(1,"Matrix is not positive definite");
	}

	det = 2.0 * log_trace(decomp, dim);
	for(i = 0; i < dim; i ++) 
	{
		decomp[i][i] = 1.0 / decomp[i][i];
		for(j = i+1; j < dim; j ++)
		{
			sum = 0.0;
			for(k = i; k < j ; k ++)
			{
				sum -= decomp[j][k] * decomp[k][i];
			}
			decomp[j][i] = sum / decomp[j][j];
		}
	}
	for(i = 0; i < dim; i ++) 
	{
		for(j = 0;j <=i; j ++)
		{
			sum = 0.0;
			for(k = 0; k < dim ; k ++)
			{
				sum += decomp[k][i] * decomp[k][j];
			}
			inv[i][j] = sum;
		}			
	}
	resort_matrix(inv, dim);

	FreeRealMatrix2(&decomp);
	return(det);
}
void outer_product(real **prod,real *vector1, real *vector2,int n) {
  int i,j;
  real *prod1;
  real vec1;
  
  for(i = 0; i < n; i ++) 
  {
	  prod1 = prod[i];
	  vec1  = vector1[i];
	  for(j = 0; j < n; j ++)
	  {
		  prod1[j] += vec1 * vector2[j];
	  }
  }
  
}  
void weight_outer_prod(real **prod,real *vector1, real *vector2,real weight,int n) {
  int i,j;
  real *prod1;
  real vec1;
  
  for(i = 0; i < n; i ++) 
  {
	  prod1 = prod[i];
	  vec1  = weight * vector1[i];
	  for(j = 0; j <= i; j ++)
	  {
		  prod1[j] += vec1 * vector2[j];
	  }
  }
}  
void resort_matrix(real **matrix,int n) {
  int i,j;
  for(i = 0; i < n; i ++)
    for(j = 0; j<i; j ++)
      matrix[j][i] = matrix[i][j];
}



void scale_matrix(real **matrix,real scale,int m,int n) {
  int i,j;
  real sc;
  real *mat1;
  
  if( scale < MIN_SCALE) {
/*    error(1,"scale too small in scale_matrix" );*/
    scale = MIN_SCALE;
  }
  sc = 1.0 / scale;
  
  for(i = 0; i < m; i ++) {
	  mat1 = matrix[i];
	  for(j = 0; j < n; j ++) {
		  mat1[j] *= sc;
	  }
  }
  
}
void unity_matrix(real **matrix,int n) {
  int i,j;
  for(i = 0; i < n; i ++)
    for(j = 0; j < n; j ++)
      matrix[i][j] = 0.0;
  for(i = 0; i < n; i ++)
    matrix[i][i] = 1.0;
}

void copy_matrix(real **matrix1,real **matrix2,int m,int n) {
  int i,j;
  for(i = 0; i < m; i ++) for(j = 0; j < n; j ++)
    matrix2[i][j] = matrix1[i][j];
}
void zero_matrix(real **matrix,int m,int n) {
  int i,j;
  for(i = 0; i < m; i ++) for(j = 0; j < n; j ++)
    matrix[i][j] = 0.0;
}
void add_matrices(real **matrix1,real **matrix2,int m,int n) {
  int i,j;
  for(i = 0; i < m; i ++)
    for(j = 0; j < n; j ++) 
      matrix1[i][j] += matrix2[i][j];
}
void weight_add_matrices(real **matrix1,real **matrix2,real weight,int m,int n) {
  int i,j;
  for(i = 0; i < m; i ++)
    for(j = 0; j < n; j ++) 
      matrix1[i][j] += weight * matrix2[i][j];
}
real trace_matrix(real **matrix,int n) {
  int i;
  real trace = 0.0;
  for(i = 0; i < n; i ++) trace += matrix[i][i];
  return trace;
}
void print_int_matrix(int **matrix,int m,int n) {
  int i,j;
  for(i = 0; i < m; i ++) {
    for(j = 0; j < n; j ++) 
      printf("%3d ",matrix[i][j] );
    printf("\n" );
  }
  printf("\n" );
}
void fprint_matrix(FILE * out, real **matrix,int m,int n) {
  int i,j;
  for(i = 0; i < m; i ++) {
    for(j = 0; j < n; j ++) 
      fprintf(out, "%3f ",matrix[i][j] );
    fprintf(out, "\n" );
  }
  fprintf(out, "\n" );
}

void print_matrix(real **matrix,int m,int n) {
  int i,j;
  for(i = 0; i < m; i ++) {
    for(j = 0; j < n; j ++) 
      printf("%3f ",matrix[i][j] );
    printf("\n" );
  }
  printf("\n" );
}

void rand_cov_matrix(real **matrix,int seed,real offset,int n) {
  int i,j;
  srand48(seed );
  for(i = 0; i < n; i ++) 
    for(j = 0; j <= i; j ++)
      matrix[i][j] = 2.0*(drand48() * offset);
  resort_matrix(matrix,n );
}


void rand_matrix(real **matrix,int seed,real offset,int m,int n) {
  int i,j;
  seed = 0;
/*  ran_seed(seed );*/
  for(i = 0; i < m; i ++) 
	  for(j = 0; j < n; j ++)
		  matrix[i][j] = (drand48()-0.5) * offset;
}

real weight_diff_inner_prod(real *vector1, real *vector2,real weight,int n) {
  int i;
  
  real error = 0.0;
  real prod  = 0.0;
   
  for(i=0; i<n; i++) {
    error = vector1[i] - vector2[i];
    prod += error * error *weight;
  }
  return prod;
}
real diff_inner_product(real *vector1, real *vector2,int n) {
  int i;
  
  real error = 0.0;
  real prod  = 0.0;
   
  for(i=0; i<n; i++) {
    error = vector1[i] - vector2[i];
    prod += error * error;
  }
  return prod;
}

void vector_product(real *output,real *vector1, real *vector2,int n) {
  int i;
  for(i=0;i<n;i++)
    output[i] = vector1[i] * vector2[i];
}

real weight_inner_prod(real *vector1, real *vector2,real weight,int n) {
  
  int i;
  real prod = 0.0;;
  
  for(i=0; i<n; i++) 
    prod += weight * vector1[i] * vector2[i];

  return prod;
}  

void init_vector_sc(real *vector,real weight,int n) {
  int i;
  for(i=0;i<n;i++)
    vector[i] = weight;
}

void copy_vector(real *vector1,real *vector2,int n) {
  int i;
  for(i=0; i<n; i++) vector2[i] = vector1[i];
}
  
void zero_vector(real *vector,int n) {
  int i;
  for(i=0;i<n;i++) vector[i] = 0.0;
}

void norm_vector(real *vector,int n) {
  int i;
  for(i=0;i<n;i++) vector[i] = 1.0/n;
}

void weight_add_vectors(real *vector1,real *vector2,real weight,int n) {
  int i;
  for(i=0;i<n;i++)
    vector2[i] += weight * vector1[i];
}
void add_vectors(real *vector1,real *vector2,int n) {
  int i;
  for(i=0;i<n;i++)
    vector2[i] += vector1[i];
}

void fprint_vector(FILE *out,real *vector,int n) {
  int i;
  for(i=0;i<n;i++) {
    fprintf(out,"%3f ",vector[i]);
  }
  fprintf(out,"\n");
}

void print_vector(real *vector,int n) {
  int i;
  for(i=0;i<n;i++) {
    printf("%3f ",vector[i]);
  }
  printf("\n");
}

void print_dvector(double *vector,int n) {
  int i;
  for(i=0;i<n;i++) {
    printf("%3f ",vector[i]);
  }
  printf("\n");
}


void check_vector(real *vector, int opdim) {
  int i;
  for(i=0;i<opdim;i++) {
   if(isnan(vector[i])) error(1,"vector is NAN");
   //   if(isinf(vector[i])) error(1,"vector is INF");
  }
}

/**********************************************************************
 *
 *	extend a vector
 */
void	extend_vector(real *in, real *out, long dim)
{
        long	i, j;
    
	/* generate linear terms */
	for (i = 0; i < dim; i++)
	    *out++ = in[i];
	
	/* generate squared terms */
	for (i = 0; i < dim; i++)
	    for (j = 0; j <= i; j++)
		*out++ = in[i]*in[j];
}
/**********************************************************************
 *
 *	add one to end of vector
 */
void    add_vector_one(real *in, long dim) 
{
    in[dim - 1] = 1.0;
}

void rand_vector(real *vector, int dim, real offset, int seed) 
{
	int i;
	seed = 0;
	for(i = 0; i < dim; i ++)
	  vector[i] = offset * (drand48()-0.5);
	
}


void rand_prob_vector(real *vector,int seed,real offset,int n) {
  int i;
  real sum;
  seed = 0;
  if(offset > 0.0) {
    for(i=0;i<n;i++) vector[i] = drand48()*offset;
    sum = trace_vector(vector,n);
    scale_vector(vector,sum,n);
  }
  else 
    norm_vector(vector,n);
}

float max_vf(float* vector, int n)
{
  float max = FLT_MIN;
  int best = -1;
  for(int i = 0; i < n; i ++){
    if(vector[i] > max){
      max = vector[i];
      best = i;
    }
  }
  if(best != -1)
    return vector[best];
  else error(1, "error in max_vf");
  return 0.0;
}
int max_index_vf(float* vector, int n)
{
  float max = FLT_MIN;
  int best = -1;
  for(int i = 0; i < n; i ++){
    if(vector[i] > max){
      max = vector[i];
      best = i;
    }
  }
  if(best != -1)
    return best;
  else error(1,"stuff up in max_index_vf");
  return -1;
}

  
      



/* matCopy:  Copy all the elements of one matrix into another */

void matCopy(real** m0, real** m1, int m, int n) {
  int i, j;

  if(m0 != m1)
    for(i = 0; i < m; i++)
      for(j = 0; j < n; j++)
	m1[i][j] = m0[i][j];
}

/* matTran:  Performs the matrix transpose operation */

void matTran(real** m0, real** m1, int m, int n) {
  int i, j;

  /* works in place for square matrices */
  if(m0 == m1) {
    if(m != n)
      error(1,"MatTrans: same memory but different dimensions!\n");

    for(i = 0; i < n; i++)
      for(j = 0; j < i; j++) {
        real tmp = m0[i][j];
        m0[i][j] = m0[j][i];
        m0[j][i] = tmp;
      }
  }
  else
    for(i = 0; i < m; i++)
      for(j = 0; j < n; j++)
        m1[j][i] = m0[i][j];
}


/* matMult: multiply a m by n matrix with a n by p to give a m by p matrix */

void matMult(real** m0, real** m1, real** m2, int m, int n, int p) {
  int i, j, k;

  for(i = 0; i < m; i++)
    for(j = 0; j < p; j++) {
      real sum = 0.0;
      
      for(k = 0; k < n; k++)
	sum += m0[i][k] * m1[k][j];
      m2[i][j] = sum;
    }
}
