#include "incs.h"
#include "net_class.h"
#include "utils.h"
#include "functions.h"

NET_Class::NET_Class(int n_batch, 
		     int num_ip, 
		     int num_op,
		     int dbg,
		     int _msteps,
		     real _tol)
  :    n_weights(num_ip * num_op),
       msteps(_msteps),
       tol(_tol),
       Net(n_batch, num_ip, num_op, dbg)
{
  weights = RealVector(n_weights);
  dweights = RealVector(n_weights);
  last_weights = RealVector(n_weights);
  for(int w = 0; w < n_weights; w ++) {
	weights[w] = 0.0;
	dweights[w] = 0.0;
	last_weights[w] = 0.0;
  }
  eta = RealVector(n_output);
  strcpy(class_str, "NET_Class");
}
  
NET_Class::~NET_Class()
{
  FreeRealVector(&weights);
  FreeRealVector(&dweights);
  FreeRealVector(&last_weights);
}

NET_Class::NET_Class(const NET_Class & old)
  :    n_weights(old.n_weights),
       msteps(old.msteps),
       tol(old.tol),
       Net(old)
{
  weights = RealVector(n_weights);
  dweights = RealVector(n_weights);
  last_weights = RealVector(n_weights);
  eta = RealVector(n_output);
  for(int i =0; i < n_weights; i ++) {
	weights[i] = old.weights[i];
  }
  strcpy(class_str, "NET_Class");
}

NET_Class::NET_Class(int n_batch, FILE* fp) 
  : Net(n_batch, fp)
{
  n_weights = n_input * n_output;
  strcpy(class_str, "NET_Class");
  weights = RealVector(n_weights);
  dweights = RealVector(n_output * n_input);
  last_weights = RealVector(n_output * n_input);
  eta = RealVector(n_output);
  for(int i = 0; i < n_output; i ++){
    for(int j = 0; j < n_input; j ++){
      float tmp;
      fscanf(fp, "%f ", &tmp);
      weights[i*n_input +j] = (real) tmp;
    }
  }  
}  

void NET_Class::randomize_weights(real offset)
{
  for(int wt_num = 0; wt_num < n_weights; wt_num ++) {
	double norm = cos(2*PI * drand48()) * sqrt(-2.0 * log(drand48()));
    weights[wt_num] = offset * norm / sqrt(n_input);
  }
}


void NET_Class::randomize_weights2(real offset)
{
  for(int wt_num = 0; wt_num < n_weights; wt_num ++) {
    weights[wt_num] = (1.0 + offset * 2.0*(get_drand() - 0.5)) * weights[wt_num]  ;

  }
}


void NET_Class::go_forward(real* in, real* out)
{
  for(int o = 0;  o < n_output;  o ++) {
    eta[o] = 0.0;
  }
  for(int op2 = 0;  op2 < n_output;  op2 ++) {
    for(int ip = 0; ip < n_input; ip ++) {
      eta[op2] += weights[op2*n_input + ip] * in[ip];
	  if(debug > 2) {
		check_arg(weights[op2*n_input +ip], "net_class : weight");
		check_arg(in[ip], "net_class : in");}
    }
  }
  log_softmax(out, eta, n_output);
  for(int op = 0;  op < n_output;  op ++) {
	out[op] = Sexp(out[op]);
	if(out[op] < VSMALL) out[op] = VSMALL;
	if(debug > 2)
	  check_arg(out[op], "net_class: out");
  }
}

void NET_Class::go_forward_log(real* in, real* out)
{
  for(int o = 0;  o < n_output;  o ++) {
    eta[o] = 0.0;
  }
  for(int op = 0;  op < n_output;  op ++) {
    for(int ip = 0; ip < n_input; ip ++) {
      eta[op] += weights[op*n_input + ip] * in[ip];
    }
  }
  log_softmax(out, eta, n_output);
  if(debug > 2)
	for(int op2 = 0;  op2 < n_output;  op2 ++) 
	  check_arg(out[op2], "net_class: log_out");
}
void NET_Class::go_forward(real* in, real* out, real* m2)
{
  for(int op = 0;  op < n_output;  op ++) {
	m2[op] = 0.0;
  }
  go_forward(in, out);
}

double NET_Class::train(int n_points, real** in, real** targ)
{
  for(int n = 0; n < n_points; n ++) {
	jpost[n] = 1.0;
  }
  inputs = in;
  targets = targ;
  double cost = train_params(n_points);
  return cost;
}

double NET_Class::train(int n_points, real** in, real** targ, real* wts)
{
  
 for(int n = 0; n < n_points; n ++) {
   jpost[n] = Sexp(wts[n]);
   if(jpost[n] < VSMALL) jpost[n] = VSMALL;
 }
  inputs = in;
  targets = targ;
 
  for(int w1 = 0; w1 < n_weights; w1 ++) 
	last_weights[w1] = weights[w1];

  double cost =  train_params(n_points, jpost);
  for(int w = 0; w < n_weights; w ++)
	dweights[w] = weights[w] - last_weights[w];
  return cost;
}

void NET_Class::write_weights(FILE* fp)
{
  fprintf(fp, "%s\n", class_str);
  fprintf(fp, "%d %d\n", n_output, n_input);
  for(int i = 0; i < n_output; i ++){
    for(int j = 0; j < n_input; j ++){
      fprintf(fp, "%f ", (float) weights[i*n_input +j]);
    }
    fprintf(fp, "\n");
  }
}

void NET_Class::read_weights(FILE* fp)
{
  int ninp;
  int noutp;
  fscanf(fp, "%s\n", class_str);
  if(strstr(class_str, "NET_Class") == NULL)
	error(1, "No class id %s found", class_str);
  fscanf(fp, "%d %d\n", &noutp, &ninp);
  for(int i = 0; i < n_output; i ++){
    for(int j = 0; j < n_input; j ++){
      float tmp;
      fscanf(fp, "%f ", &tmp);
      weights[i*n_input +j] = (real) tmp;
    }
  }
}
void NET_Class::get_gradient(double* grad) {
  for(int n = 0; n < n_weights; n ++) {
    grad[n] = 0.0;
  }
  for(int point = 0; point < batch_size; point ++) {
    go_forward(inputs[point], tmp_op);
	real wt = jpost[point];
	//	if(wt > 0.0) {
	  int wt_num = 0;
	  for(int op = 0; op < n_output; op ++) {
		double err = (tmp_op[op] - targets[point][op]);
		double wt_err = wt * err;
		for(int ip =0; ip < n_input; ip ++) {
		  grad[wt_num] +=  wt_err * inputs[point][ip];
		  wt_num ++;
		}
		//	  }
    }
  }
}
void NET_Class::get_gradient(double* grad, double* fn) {
 for(int n = 0; n < n_weights; n ++) {
    grad[n] = 0.0;
  }
  *fn = 0.0;
  for(int point = 0; point < batch_size; point ++) {
    go_forward(inputs[point], tmp_op);
    int wt_num = 0;
	real wt = jpost[point];
	//	if(wt > 0.0) {
	  for(int op = 0; op < n_output; op ++) {
		double err = (tmp_op[op] - targets[point][op]);
		double err2 = err * wt;
		for(int ip =0; ip < n_input; ip ++) {
		  grad[wt_num] +=  err2 *  inputs[point][ip];
		  wt_num ++;
		}
		(*fn) -= wt * targets[point][op] * Slog(tmp_op[op]);
	  }
	  //	}
  }
}

double NET_Class::get_function() {
  double fn = 0.0;
  for(int point = 0; point < batch_size; point ++) {
    go_forward(inputs[point], tmp_op);
	double wt = jpost[point];
	//	if(wt > 0.0) {
	  for(int op = 0; op < n_output; op ++) {
		fn -= wt * targets[point][op] * Slog(tmp_op[op]);
	  }
	  //	}
  }
  return fn;
}
void NET_Class::compute_hessian(int n_points, 
				  real** in, 
				  real** out, 
				  real* wts,
				  real*** hess)
{
  for(int o2 = 0; o2 < n_output; o2 ++) {
    for(int i = 0; i < n_input; i ++) {
      for(int i2 = 0; i2 < n_input; i2 ++) {
		hess[o2][i][i2] = 0.0;
      }
    }
  }
  for(int n = 0; n < n_points; n ++) {
    for(int o = 0; o < n_output; o ++) {
      real w = wts[n] * out[n][o] * (1.0 - out[n][o]);
      for(int i = 0; i < n_input; i ++) {
	for(int i2 = 0; i2 < n_input; i2 ++) {
	  hess[o][i][i2] += w  * in[n][i] * in[n][i2];
	}
      }
    }
  }
}

double NET_Class::get_joint(real* in, real* targ)
{
  double joint = 0.0;
  go_forward_log(in, tmp_op);
  for(int op = 0; op < n_output; op ++) {
	joint += targ[op] * tmp_op[op];
  }
  return (joint);
}

void NET_Class::update_hyperparameters(int n_points, real** in, 
									   real** targ, real* wts) 
{
}

void NET_Class::init(int n_points, real** in, 
					 real** targ, real* wts) {

  train(n_points, in, targ, wts);
}
real NET_Class::get_sum_jposts(int n_points, real* wts) {
  real sum = LZERO;
  for (int n = 0; n < n_points; n ++) {
	sum = LAdd(sum, wts[n]);
  }
  return sum;
}

real NET_Class::get_evidence(int n_points, real** in, real** targ,
								   real* wts) {
  return 0.0;
}
