#include "incs.h"
#include "net_regress.h"
#include "net.h"
#include "utils.h"
#include "functions.h"

NET_Regress::NET_Regress(int n_batch, 
						 int num_ip, 
						 int num_op,
						 real _lambda,
						 real _prior_var,
						 int dbg,
						 int _msteps,
						 real _tol
						 )
  : n_weights(num_ip * num_op),
    msteps(_msteps),
    tol(_tol),
    Net(n_batch, num_ip, num_op, dbg),
    lambda(_lambda),
    prior_var(_prior_var)
{
  weights = RealVector(n_weights);
  dweights = RealVector(n_weights);
  last_weights = RealVector(n_weights);
  for(int w = 0; w < n_weights; w ++) {
	weights[w] = 0.0;
	dweights[w] = 0.0;
	last_weights[w] = 0.0;
  }
  var_acc = RealVector(n_output);
  variance = RealVector(n_output);

  for(int i = 0; i < n_output; i ++) {
	variance[i] = prior_var/lambda;
  }
  inv_variance = RealVector(n_output);
  for(int j = 0; j < n_output; j ++) {
	inv_variance[j] = 1.0 / variance[j];
  }
  strcpy(class_str, "NET_Regress");
}

NET_Regress::NET_Regress(const NET_Regress & old)
  : n_weights(old.n_weights),
    msteps(old.msteps),
    tol(old.tol),
    Net(old),
    lambda(old.lambda),
    prior_var(old.prior_var)
{
  weights = RealVector(n_weights);
  dweights = RealVector(n_output * n_input);
  last_weights = RealVector(n_output * n_input);

  for(int i = 0; i < n_weights; i ++) {
	weights[i] = old.weights[i];
  }
  var_acc = RealVector(n_output);
  for(int j = 0; j < n_output; j ++) {
	var_acc[j] = old.var_acc[j];
  }
  variance = RealVector(n_output);
  for(int k = 0; k < n_output; k ++) {
    variance[k] = old.variance[k];
  }
  inv_variance = RealVector(n_output);
  for(int l = 0; l < n_output; l ++) {
    inv_variance[l] = old.inv_variance[l];
  }
  strcpy(class_str, "NET_Regress");
}

NET_Regress::NET_Regress(int n_batch, FILE* fp) 
: Net(n_batch, fp)
{
  strcpy(class_str, "NET_Regress");
  weights = RealVector(n_output * n_input);
  dweights = RealVector(n_output * n_input);
  last_weights = RealVector(n_output * n_input);

  var_acc = RealVector(n_output);
  variance = RealVector(n_output);
  inv_variance = RealVector(n_output);

  for(int i = 0; i < n_output; i ++){
    for(int j = 0; j < n_input; j ++){
      float tmp;
      fscanf(fp, "%f ", &tmp);
      weights[i*n_input +j] = (real) tmp;
    }
  }  
  for(int j = 0; j < n_output; j ++){
    float tmp;
    fscanf(fp, "%f ", &tmp);
    variance[j] = (real) tmp;
  }

  for(int k = 0; k < n_output; k ++) {
	inv_variance[k] = 1.0 / variance[k];
  }
}  
  
NET_Regress::~NET_Regress()
{
  FreeRealVector(&weights);
  FreeRealVector(&dweights);
  FreeRealVector(&last_weights);
  FreeRealVector(&var_acc);
  FreeRealVector(&variance);
  FreeRealVector(&inv_variance);
}

void NET_Regress::randomize_weights(real offset)
{
  for(int wt_num = 0; wt_num < n_weights; wt_num ++) {
    weights[wt_num] = offset * 2.0*(get_drand() - 0.5);
  }
  /*  for(int wt_num = 0; wt_num < n_weights; wt_num ++) {
	double norm = cos(2*PI * drand48()) * sqrt(-2.0 * log(drand48()));
    weights[wt_num] = offset * norm / sqrt(n_input);
  }*/
}

void NET_Regress::randomize_weights2(real offset)
{
  for(int wt_num = 0; wt_num < n_weights; wt_num ++) {
	weights[wt_num] =  (1.0 + offset * 2.0*(get_drand() - 0.5)) * weights[wt_num]  ;
  }
}

void NET_Regress::go_forward(real* in, real* out)
{
  for(int op1 = 0;  op1 < n_output;  op1 ++) {
    out[op1] = 0.0;
  }
  for(int op = 0;  op < n_output;  op ++) {
    for(int ip = 0; ip < n_input; ip ++) {
      out[op] += weights[op*n_input + ip] * in[ip];
    }
  }
}

void NET_Regress::go_forward_log(real* in, real* out)
{
  for(int op1 = 0;  op1 < n_output;  op1 ++) {
    out[op1] = 0.0;
  }
  for(int op = 0;  op < n_output;  op ++) {
    for(int ip = 0; ip < n_input; ip ++) {
      out[op] += weights[op*n_input + ip] * in[ip];
    }
  }
}
void NET_Regress::go_forward(real* in, real* out, 
							 real* m2)
{
   for(int op1 = 0;  op1 < n_output;  op1 ++) {
     out[op1] = 0.0;
   }
   for(int op2 = 0;  op2 < n_output;  op2 ++) {
     for(int ip = 0; ip < n_input; ip ++) {
       out[op2] += weights[op2*n_input + ip] * in[ip];
     }
   }
   for(int op = 0; op < n_output; op ++) {
     m2[op] = variance[op] + (out[op] * out[op]);
   }
}

void NET_Regress::read_weights(FILE* fp) {
  int ninp;
  int noutp;
  fscanf(fp, "%s\n", class_str);
  if(strstr(class_str, "NET_Regress") == NULL)
	error(1, "No class id %s found", class_str);
  fscanf(fp, "%d %d\n", &noutp, &ninp);
  for(int i = 0; i < n_output; i ++){
    for(int j = 0; j < n_input; j ++){
      float tmp;
      fscanf(fp, "%f ", &tmp);
      weights[i*n_input +j] = (real) tmp;
    }
  }
  for(int j = 0; j < n_output; j ++){
    float tmp;
    fscanf(fp, "%f ", &tmp);
    variance[j] = (real) tmp;
  }
 for(int k = 0; k < n_output; k ++) {
	inv_variance[k] = 1.0 / variance[k];
  }
}

void NET_Regress::write_weights(FILE* fp)
{
  fprintf(fp, "%s\n", class_str);
  fprintf(fp, "%d %d\n", n_output, n_input);
  for(int i = 0; i < n_output; i ++){
    for(int j = 0; j < n_input; j ++){
      fprintf(fp, "%f ", (float) weights[i*n_input +j]);
    }
    fprintf(fp, "\n");
  }
  for(int j = 0; j < n_output; j ++){
    fprintf(fp, "%f ", (float) variance[j]);
  }
  fprintf(fp, "\n");
}

void NET_Regress::compute_hessian(int n_points, 
								  real** in, 
								  real* wts,
								  real*** hess)
{
  for(int o = 0; o < n_output; o ++) {
    for(int i = 0; i < n_input; i ++) {
      for(int i2 = 0; i2 < n_input; i2 ++) {
	hess[o][i][i2] = 0.0;
      }
    }
  }
  for(int n = 0; n < n_points; n ++) {
    for(int o2 = 0; o2 < n_output; o2 ++) {
      real w = wts[n] * inv_variance[o2];
      for(int i = 0; i < n_input; i ++) {
		for(int i2 = 0; i2 < n_input; i2 ++) {
		  hess[o2][i][i2] += w  * in[n][i] * in[n][i2];
		}
      }
    }
  }
}

double NET_Regress::train(int n_points, real** in, real** targ)
{
  for(int n = 0; n < n_points; n ++) {
    jpost[n] = 1.0;
  }
  inputs = in;
  targets = targ;
  // this is a hack to get round the / by variance in
  // train_params...
  for(int o = 0; o < n_output; o ++) {
	variance[o] = 1.0;
  }
  double cost = train_params(n_points);
  go_forward_batch(n_points, in, tmp_op_vec);
  update_variance(n_points, tmp_op_vec, targ);
  return cost;
}

double NET_Regress::train(int n_points, real** in, real** targ, real* wts)
{
  for(int n = 0; n < n_points; n ++) {
    jpost[n] = Sexp(wts[n]);
	if(jpost[n] < VSMALL) jpost[n] = VSMALL;
  }
  inputs = in;
  targets = targ;
  
  for(int w1 = 0; w1 < n_weights; w1 ++) 
	last_weights[w1] = weights[w1];
  double cost =  train_params(n_points, jpost);
  for(int w = 0; w < n_weights; w ++)
	dweights[w] = weights[w] - last_weights[w];
  return cost;
}
void NET_Regress::init(int n_points, 
					   real** in, 
					   real** targ,
					   real* wts)
{
  inputs = in;
  targets = targ;

  for(int n = 0; n < n_points; n ++) {
    jpost[n] = Sexp(wts[n]);
	if(jpost[n] < VSMALL) jpost[n] = VSMALL;
  }
  go_forward_batch(n_points, in, tmp_op_vec);
  update_variance(n_points, tmp_op_vec, targ, jpost);
  double cost =  train_params(n_points, jpost);
  go_forward_batch(n_points, in, tmp_op_vec);
  update_variance(n_points, tmp_op_vec, targ, jpost);
  if(debug > 0)
	printf("cost = %f\n", cost);
}
double NET_Regress::get_joint(real* in, real* target)
{
  // each output is modelled by a Gaussian
  // with mean (out[i]-targ[i]) and variance[i]
  real arg = 0.0;
  go_forward(in, tmp_op);
  for(int i = 0; i < n_output; i ++) {
    real error1 = tmp_op[i] - target[i];
    arg += error1 * error1 * inv_variance[i];
	arg += Slog(variance[i]);
  }
  double joint = -arg;
  return joint;
}

void NET_Regress::update_variance(int n_points, 
								  real** out, 
								  real** targ,
								  real* wts) {

  for(int k = 0; k < n_output; k ++) {
    var_acc[k] = 0.0;
  }
  double sum_wts = 0.0;
  for(int point2 = 0; point2 < n_points; point2 ++){
	double wt = wts[point2];
    sum_wts += wt;
    for(int op =0; op < n_output; op ++) {
      real v = (targ[point2][op] - out[point2][op]);
      var_acc[op] += wt * v*v;
    }
  }
  for(int j = 0; j < n_output; j ++) {
	variance[j] = (var_acc[j] + lambda * prior_var) / (sum_wts + lambda);
  }

  for(int i = 0; i < n_output; i ++) {
	inv_variance[i] = 1.0 / variance[i];
  }
  if(debug > 0){
	for(int l = 0; l < n_output; l ++) {
	  printf("variance = %f %f %f\n", variance[l], sum_wts, var_acc[l]);
    }
  }
}

void NET_Regress::update_variance(int n_points, 
								  real** out, 
								  real** targ) {
  for(int k = 0; k < n_output; k ++) {
    var_acc[k] = 0.0;
  }
  for(int point2 = 0; point2 < n_points; point2 ++){
    for(int op =0; op < n_output; op ++) {
      real v = (targ[point2][op] - out[point2][op]);
      var_acc[op] += v * v;
    }
  }
  for(int j = 0; j < n_output; j ++) {
	variance[j] = (var_acc[j]) / n_points;
  }
  for(int i = 0; i < n_output; i ++) {
	inv_variance[i] = 1.0 / variance[i];
  }
  if(debug > 0 ) {
	for(int l = 0; l < n_output; l ++) {
	  printf("variance = %f \n", variance[l]);
	}
  }
}
void NET_Regress::update_hyperparameters(int n_points, 
										 real** in,
										 real** targ, real* wts) {
  inputs = in;
  targets = targ;
  for(int o = 0; o < n_points; o ++) {
	jpost[o] = Sexp(wts[o]);
	if(jpost[o] < VSMALL) jpost[o] = VSMALL;
  }
  go_forward_batch(n_points, in, tmp_op_vec);
  update_variance(n_points, tmp_op_vec, targ, jpost);
}

void NET_Regress::get_gradient(double* grad) {
  for(int n = 0; n < n_weights; n ++) {
    grad[n] = 0.0;
  }
  for(int point = 0; point < batch_size; point ++) {
    go_forward(inputs[point], tmp_op);
    int wt_num = 0;
	real wt = jpost[point];
	//	if(wt > 0.0) {
	  for(int op = 0; op < n_output; op ++) {
		double err = wt * (tmp_op[op] - targets[point][op]);
		for(int ip =0; ip < n_input; ip ++) {
		  grad[wt_num] +=  err * inputs[point][ip];
		  wt_num ++;
		  //		}
		}
	}
  }
  int wt_num = 0;
  for(int op = 0; op < n_output; op ++) {
	for(int ip =0; ip < n_input; ip ++) {
	  grad[wt_num] *= inv_variance[op];
	  wt_num ++;
	}
  }
}
void NET_Regress::get_gradient(double* grad, double* fn) {
 for(int n = 0; n < n_weights; n ++) {
    grad[n] = 0.0;
  }
  *fn = 0.0;
  for(int point = 0; point < batch_size; point ++) {
    go_forward(inputs[point], tmp_op);
    int wt_num = 0;
	real wt = jpost[point];
//	if(wt > 0.0) {
	  for(int op = 0; op < n_output; op ++) {
		double err = (tmp_op[op] - targets[point][op]);
		double err2 = err * wt;
		for(int ip =0; ip < n_input; ip ++) {
		  grad[wt_num] +=  err2 * inputs[point][ip];
		  wt_num ++;
		}
		(*fn) += wt * err * err;
//	  }
	}
  }
  int wt_num = 0;
  for(int op = 0; op < n_output; op ++) {
	for(int ip =0; ip < n_input; ip ++) {
	  grad[wt_num] *= inv_variance[op];
	  wt_num ++;
	}
  }
}

double NET_Regress::get_function() {
  double fn = 0.0;
  for(int point = 0; point < batch_size; point ++) {
	real wt = jpost[point];
	//	if(wt > 0.0) {
	go_forward(inputs[point], tmp_op);
	for(int op = 0; op < n_output; op ++) {
	  double err = tmp_op[op] - targets[point][op] ;
	  fn += wt * err * err;
	  //	  }
	}
  }
  return fn;
}
real NET_Regress::get_sum_jposts(int n_points, real* wts) {
  real sum = LZERO;
  for (int n = 0; n < n_points; n ++) {
	sum = LAdd(sum, wts[n]);
  }
  return sum;
}

real NET_Regress::get_evidence(int n_points, real** in, real** targ,
								   real* wts) {
  return 0.0;
}
